(*
 * Copyright (C) 2008 Jean-Christophe Filliatre
 * Copyright (C) 2008, 2009 Laurent Hubert (CNRS)
 *
 * This software is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public License
 * version 2.1, with the special exception on linking described in file
 * LICENSE.

 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 *)

(** Maps over integers implemented as Patricia trees.

    This implementation follows Chris Okasaki and Andrew Gill's paper
    {e Fast Mergeable Integer Maps}.*)

(** The following signature is similar to [Map.S with type key = int].
    The documentation is only given for function that differs from
    [Map.S with type key = int].

    The module {!Ptset} is based on the same data-structure.

*)

module type S = sig
  type +'a t
  type key = int

  val empty : 'a t
  val is_empty : 'a t -> bool

  val add : ?merge:('a -> 'a -> 'a) -> int -> 'a -> 'a t -> 'a t
  (** [add ~merge:f k d m] returns a map containing the same bindings as
      [m], plus a binding of [k] to [d].  If [k] was already bound to
      [d'] in [m], then the value [f d' d] is added instead of [d].  If
      no merge function is specified, then the previous bindings is
      simply discard.
   *)

  val modify : int -> ('a option -> 'a) -> 'a t -> 'a t
  val find : int -> 'a t -> 'a

  val findi_element : (int -> 'a -> bool) -> 'a t -> int * 'a
  (** [findi_element f t] returns the first couple (index,element) in map [t] for which [f ind elt], with [ind] an index and [elt] the corresponding element, returns true.

    @raise Not_found if [t] is empty or if [f] returns false for all
    elements.*)

  val find_element : ('a -> bool) -> 'a t -> 'a
  val remove : int -> 'a t -> 'a t
  val mem : int -> 'a t -> bool
  val iter : (int -> 'a -> unit) -> 'a t -> unit

  val iter_ordered :
    (int * 'a -> int * 'a -> int) -> (int -> 'a -> unit) -> 'a t -> unit

  val map : ('a -> 'b) -> 'a t -> 'b t
  val mapi : (int -> 'a -> 'b) -> 'a t -> 'b t
  val fold : (int -> 'a -> 'b -> 'b) -> 'a t -> 'b -> 'b

  val fold_ordered :
    (int * 'a -> int * 'a -> int) -> (int -> 'a -> 'b -> 'b) -> 'a t -> 'b -> 'b

  val compare : ('a -> 'a -> int) -> 'a t -> 'a t -> int
  val equal : ('a -> 'a -> bool) -> 'a t -> 'a t -> bool

  val merge : ('a -> 'a -> 'a) -> 'a t -> 'a t -> 'a t
  (** [merge f m1 m2] returns a map that binds every key [k] that is
			either bound in [m1] or/and in [m2]. 
			If [k] was bound to [d1] in [m1] and [d2] in [m2], it is now bound to [f d1 d2]. 
			If [k] was bound to [d1] in [m1] but is not bound in [m2], it is still
			bound to [d1].
			If [k] was bound to [d2] in [m2] but is not bound in [m1], it is still
			bound to [d2].
			Precondition on [f]: if [d1] equals [d2], [f d1 d2] is supposed to return [d1].

   *)

  val merge_first : 'a t -> 'a t -> 'a t
  (** [merge_first m1 m2] is the same as [merge (fun a _ -> a) m1 m2] but it
      reuses more data from the first map.*)

  val diff : ('a -> 'a -> bool) -> 'a t -> 'a t -> 'a t

  val choose_and_remove : 'a t -> int * 'a * 'a t
  (** [choose_and_remove t] returns (i,d,t') such that [t'] equals to
      [remove i t] and [d] equals to [find i t].

      @raise Not_found if [t] is empty.
      @author Laurent Hubert*)

  val inter : 'a t -> 'a t -> 'a t
  (** [inter m1 m2] returns a map with all the bindings [(k,d)] of [m1] such
      that [mem k m2]. *)

  (* [inter_map2 f m1 m2] returns a map that binds only the keys [k]
     that are binded both in [m1] and [m2]. The associated value is
     [f (find k m1) (find k m2)]. [f] must satisfy [f x x = x] for all
     [x].

     @author David Pichardie *)
  val inter_map2 : ('a -> 'a -> 'a) -> 'a t -> 'a t -> 'a t

  val keys_subset : 'a t -> 'a t -> bool
  (** [keys_subset m1 m2] returns [true] if the set of keys in [m1] is a subset of
      the set of keys in [m2], [false] otherwise. *)

  val subset : ('a -> 'a -> bool) -> 'a t -> 'a t -> bool
  (** [subset sub m1 m2] returns [true] if the set of keys in [m1] is a subset of
      the set of keys in [m2], and if for all value [v1] in [m1], and [v2] in [m2],
      at the same position, [sub v1 v2=true] holds,
      [false] otherwise. *)

  val cardinal : 'a t -> int
  val exists : (int -> 'a -> bool) -> 'a t -> bool
  val filter : ('a -> bool) -> 'a t -> 'a t
  val filteri : (int -> 'a -> bool) -> 'a t -> 'a t
  val filter_map : ('a -> 'b option) -> 'a t -> 'b t
  val partition : ('a -> bool) -> 'a t -> 'a t * 'a t

  val elements : 'a t -> (int * 'a) list
  (** [elements m] return the unsorted list of bindings of [m].  *)
end

include S
